/***************************************************************************
 *  InfoType.cs
 *
 *  Copyright (C) 2005 Novell
 *  Written by Oscar Forero (oforero@novell.com)
 ****************************************************************************/

/*  THIS FILE IS LICENSED UNDER THE MIT LICENSE AS OUTLINED IMMEDIATELY BELOW: 
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a
 *  copy of this software and associated documentation files (the "Software"),  
 *  to deal in the Software without restriction, including without limitation  
 *  the rights to use, copy, modify, merge, publish, distribute, sublicense,  
 *  and/or sell copies of the Software, and to permit persons to whom the  
 *  Software is furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in 
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 *  DEALINGS IN THE SOFTWARE.
 */

using System;
using System.Collections;
using System.Data;
using System.Reflection;

using Mono.Unix;

namespace Banshee.Base
{
    public class InfoTypeFactory 
    {
        private Hashtable byId = new Hashtable();
       // private Hashtable byName = new Hashtable(new CaseInsensitiveHashCodeProvider(), new CaseInsensitiveComparer());
        private SortedList byName = new SortedList(new CaseInsensitiveComparer());
		private int newId = -2;
		private ConstructorInfo constructor;
		private string id_field;
		
		public ICollection Instances {
			get { return byId.Values; }
		}		
		
		public InfoTypeFactory(Type type, string id_field)
		{
			this.id_field = id_field;
			constructor = type.GetConstructor(new Type[] {typeof(int), typeof(string)});
		}

		private InfoType GetConcreteInstance(int id, string name)
		{
			return (InfoType) constructor.Invoke(new object[] {id, name});
		}

		public InfoType GetInstance(int id, string name)
		{
			//name = name.Trim(); //Still needed?
			InfoType instance = GetConcreteInstance(id, name);
			byId.Add(id, instance);
			byName.Add(name, instance);
			return instance;
		}

		public InfoType GetById(int id) 
		{
			return byId[id] as InfoType;
		}
		
		public InfoType GetByName(string name) 
		{
			//name = name.Trim();
			InfoType type =  byName[name] as GenreInfo;
			return (type == null) ? GetInstance(newId--, name) : type;
		}
		
		public bool Exist(int id) 
		{
			return byId.ContainsKey(id);
		}

		public bool Exist(string name) 
		{
			return byName.ContainsKey(name);
		}

		public int Compare(InfoType o1, InfoType o2)
		{
			return (o1 == o2) ? 0 : byName.IndexOfKey(o1.Name) - byName.IndexOfKey(o2.Name); 	
		} 
		
		internal void LoadFromDatabaseReader(IDataReader reader)
		{
			try {
				int id = Convert.ToInt32(reader[id_field]);
    	        if(!Exist(id)) {
					GetInstance(id, reader["Name"] as string);
				}
			}
			catch(Exception e) {
				Console.WriteLine(e.Message);
			}
		}
	}
	
    abstract public class InfoType 
	{ 
		private int id;
		private string name;

		protected InfoType(int id, string name) 
		{
			this.id = id;
			this.name = name;
		}

		public int ID {
			get {
				return id;
			}
		}

		public string Name {
			get {
				return name;
			}
		}
		
		public int Compare(InfoType o)
		{
			if(id == -1) { 
				return -1; 
			} else {
				return Factory.Compare(this, o);
			}
		}

		virtual public bool IsType(InfoType type)
		{
			return id == -1 || this == type;
		}
		
		abstract public InfoType GetInstanceById(int id);
		abstract protected InfoTypeFactory Factory { get; }

	}

    public class GenreInfo : InfoType
    {
		private static InfoTypeFactory factory = new InfoTypeFactory(typeof(GenreInfo), "GenreID");
		private static GenreInfo all;
		private static GenreInfo unknown;
	
		static GenreInfo() {
			all = (GenreInfo) factory.GetInstance(-1, Catalog.GetString("All"));
			unknown = (GenreInfo) factory.GetInstance(0, Catalog.GetString("Unknown Genre"));
		}	
	
		public static GenreInfo All {
			get { return all; }
		}
	
		public static GenreInfo Unknown {
			get { return unknown; }
		}

		public static void LoadFromDatabaseReader(IDataReader reader)
		{
			factory.LoadFromDatabaseReader(reader);
		}

        public static GenreInfo GetById(int id) {
            return (GenreInfo) factory.GetById(id); 
        }

        public static GenreInfo GetByName(string name) {
            return (GenreInfo) factory.GetByName(name); 
        }
		
		public static ICollection Genres {
			get { return factory.Instances; }
		}

		private Hashtable artists;
		private Hashtable albums;

		private Hashtable Artists {
			get 
			{
				if(null==artists) 
				{
					artists = new Hashtable();
					Globals.Library.LoadTableWithInfo(artists, ArtistInfo.All, "ArtistID", "GenreID", ID);
				}
				return artists;
			}
		}
		
		private Hashtable Albums {
			get 
			{
				if(null==albums) 
				{
					albums = new Hashtable();
					Globals.Library.LoadTableWithInfo(albums, AlbumInfo.All, "AlbumID", "GenreID", ID);
				}
				return albums;
			}
		}
		
		public bool ContainsArtist(ArtistInfo artist)
		{
			return (ID == -1) || Artists.ContainsKey(artist.ID);
		}

		public bool ContainsAlbum(AlbumInfo album)
		{
			return (ID == -1) || Albums.ContainsKey(album.ID);
		}

		public bool IsType(TrackInfo track)
		{
			return base.IsType(track.Genre);
		}

		override public InfoType GetInstanceById(int id)
		{
			return GenreInfo.GetById(id);
		}

		override protected InfoTypeFactory Factory { 
			get { return GenreInfo.factory; }
		}

        public GenreInfo(int id, string name) : base(id, name) {}
	}

    public class ArtistInfo : InfoType
    {
		private static InfoTypeFactory factory = new InfoTypeFactory(typeof(ArtistInfo), "ArtistID");
		private static ArtistInfo all;
		private static ArtistInfo unknown;

        static ArtistInfo() {
            all = (ArtistInfo) factory.GetInstance(-1, Catalog.GetString("All"));
            unknown = (ArtistInfo) factory.GetInstance(0, Catalog.GetString("Unknown Artist"));
        }

		public static ArtistInfo All {
			get { return all; }
		}
	
		public static ArtistInfo Unknown {
			get { return unknown; }
		}

		public static void LoadFromDatabaseReader(IDataReader reader)
		{
			factory.LoadFromDatabaseReader(reader);
		}

        public static ArtistInfo GetById(int id) {
            return (ArtistInfo) factory.GetById(id); 
        }

        public static ArtistInfo GetByName(string name) {
            return (ArtistInfo) factory.GetByName(name); 
        }

		public static ICollection Artists {
			get { return factory.Instances; }
		}
		
		private Hashtable albums;

		private Hashtable Albums {
			get 
			{
				if(null==albums) 
				{
					albums = new Hashtable();
					Globals.Library.LoadTableWithInfo(albums, AlbumInfo.All, "AlbumID", "ArtistID", ID);
				}
				return albums;
			}
		}
		
		public bool ContainsAlbum(AlbumInfo album)
		{
			return (ID == -1) || Albums.ContainsKey(album.ID);
		}

		public bool IsType(TrackInfo track)
		{
			return base.IsType(track.Artist);
		}

		override public InfoType GetInstanceById(int id)
		{
			return ArtistInfo.GetById(id);
		}

		override protected InfoTypeFactory Factory { 
			get { return ArtistInfo.factory; }
		}

        public ArtistInfo(int id, string name) : base(id, name) {}
    }

    public class AlbumInfo : InfoType
    {
		private static InfoTypeFactory factory = new InfoTypeFactory(typeof(AlbumInfo), "AlbumID");
		private static AlbumInfo all;
		private static AlbumInfo unknown;

        static AlbumInfo() {
            all = (AlbumInfo) factory.GetInstance(-1, Catalog.GetString("All"));
            unknown = (AlbumInfo) factory.GetInstance(0, Catalog.GetString("Unknown Album"));
        }

		public static AlbumInfo All {
			get { return all; }
		}
	
		public static AlbumInfo Unknown {
			get { return unknown; }
		}

		public static void LoadFromDatabaseReader(IDataReader reader)
		{
			factory.LoadFromDatabaseReader(reader);
		}

        public static AlbumInfo GetById(int id) {
            return (AlbumInfo) factory.GetById(id); 
        }

        public static AlbumInfo GetByName(string name) {
            return (AlbumInfo) factory.GetByName(name); 
        }

		public static ICollection Albums {
			get { return factory.Instances; }
		}

		public bool IsType(TrackInfo track)
		{
			return base.IsType(track.Album);
		}

		override public InfoType GetInstanceById(int id)
		{
			return AlbumInfo.GetById(id);
		}

		override protected InfoTypeFactory Factory { 
			get { return AlbumInfo.factory; }
		}

        public AlbumInfo(int id, string name) : base(id, name) {}
    }


}
